/*
 *	cook - file construction tool
 *	Copyright (C) 1990, 1991, 1992, 1993, 1994 Peter Miller.
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * MANIFEST: functions to perform lexical analysis on cookbooks
 *
 * This file contains the lexical analyser for the cookbook parser.
 * A known bug is that it processes # control lines even within comments.
 * Another bug is that comments vanish: they should be replaced by a single
 * space.
 */

#include <ctype.h>
#include <ac/stddef.h>
#include <ac/string.h>
#include <ac/stdlib.h>
#include <stdio.h>

#include <error.h>
#include <expr.h>
#include <hashline.h>
#include <lex.h>
#include <mem.h>
#include <option.h>
#include <s-v-arg.h>
#include <stmt.h>
#include <symtab.h>
#include <trace.h>
#include <word.h>
#include <parse.gen.h>	/* must be last */
#include <hashline.gen.h> /* must be last */


/*
 * Meta types
 */
#define NORMAL	1
#define SLOSHED	2
#define WHITE	3

/*
 */
#define BEFORE	(1<<0)
#define AFTER	(1<<1)
#define SINGLE	(1<<2)

#define HASHLINE_ESCAPE ('#' ^ 0x80)
#define UNQUOTED_WORD 32767
#define EOLN 32766


typedef struct meta_ty meta_ty;
struct meta_ty
{
	int		m_type;		/* meta character "type" info	*/
	int		m_char;		/* meta character actual value	*/
	int		m_flag;		/* meta character flags		*/
};

typedef struct lex_ty lex_ty;
struct lex_ty
{
	string_ty	*l_name;	/* name of file being analysed	*/
	FILE		*l_file;	/* open file structure of the file */
	int		l_line;		/* the line number we are up to	*/
	meta_ty		l_mback;	/* backup for meta chars	*/
	lex_ty		*l_chain;	/* file this one is an insert from */
	short		l_bol;
	wlist		pending_includes;
};

typedef struct stracc stracc;
struct	stracc
{
	size_t		sa_max;		/* size of string accum buffer	*/
	size_t		sa_len;		/* number of chars accumulated	*/
	char		*sa_buf;	/* the string accum buffer	*/
	int		sa_inuse;
};

static	lex_ty		*root;		/* root of insert list		*/
static	stracc		saroot;		/* root of string accum list	*/
static	int		errcnt;		/* count of errors to date	*/
static	int		mode;		/* what lex mode we are in	*/
static	int		state;
static	int		passing;
static	int		catted;
static	symtab_ty	*parse_symtab;
static	symtab_ty	*hash_symtab;


/*
 * NAME
 *	lex_initialize - look for keywords
 *
 * SYNOPSIS
 *	int lex_initialize(void);
 *
 * DESCRIPTION
 *	The lex_initialize function adds all the keywords to the symbol table.
 *
 * CAVEAT
 *	The keywords are intentionally case sensitive.
 *	Assumes that str_initialize has already been called.
 */

void
lex_initialize()
{
	typedef struct keyword_ty keyword_ty;
	struct keyword_ty
	{
		char		*k_name;
		int		k_token;
	};

	static keyword_ty parse_keyword[] =
	{
		{ ":",			COLON,			},
		{ "=",			EQUALS,			},
		{ "[",			LBRAK,			},
		{ "]",			RBRAK,			},
		{ "data",		DATA,			},
		{ "else",		ELSE,			},
		{ "fail",		FAIL,			},
		{ "if",			IF,			},
		{ "loop",		LOOP,			},
		{ "set",		SET,			},
		{ "then",		THEN,			},
		{ "{"/*}*/,		LBRACE,			},
		{ /*{*/"}",		RBRACE,			},
		{ ";",			SEMICOLON,		},
		{ "loopstop",		LOOPSTOP,		},
		{ "unsetenv",		UNSETENV,		},
	};
	static keyword_ty hash_keyword[] =
	{
		{ "[",			LBRAK,			},
		{ "]",			RBRAK,			},
		{ "elif",		HASH_ELIF,		},
		{ "else",		HASH_ELSE,		},
		{ "endif",		HASH_ENDIF,		},
		{ "if",			HASH_IF,		},
		{ "ifdef",		HASH_IFDEF,		},
		{ "ifndef",		HASH_IFNDEF,		},
		{ "include",		HASH_INCLUDE,		},
		{ "include-cooked",	HASH_INCLUDE_COOKED,	},
		{ "include-cooked-nowarn", HASH_INCLUDE_COOKED2, },
		{ "pragma",		HASH_PRAGMA,		},
	};
	keyword_ty	*kp;

	trace(("init_reserved()\n{\n"/*}*/));
	passing = 1;

	if (!parse_symtab)
		parse_symtab = symtab_alloc(SIZEOF(parse_keyword));
	for (kp = parse_keyword; kp < ENDOF(parse_keyword); ++kp)
	{
		string_ty	*s;

		s = str_from_c(kp->k_name);
		symtab_assign(parse_symtab, s, &kp->k_token);
		str_free(s);
	}

	if (!hash_symtab)
		hash_symtab = symtab_alloc(SIZEOF(hash_keyword));
	for (kp = hash_keyword; kp < ENDOF(hash_keyword); ++kp)
	{
		string_ty	*s;

		s = str_from_c(kp->k_name);
		symtab_assign(hash_symtab, s, &kp->k_token);
		str_free(s);
	}
	trace((/*{*/"}\n"));
}


/*
 * NAME
 *	lex_open - open a file for lexical analysis
 *
 * SYNOPSIS
 *	void lex_open(string_ty *filename);
 *
 * DESCRIPTION
 *	Lex_open opens a file for lexical analysis
 *	(it is also used to open include files).
 *
 * RETURNS
 *	void
 */

void
lex_open(filename)
	string_ty	*filename;
{
	lex_ty		*new;

	trace(("lex_open(filename = %08lX)\n{\n"/*}*/, filename));
	trace_string(filename->str_text);
	if (!root)
		hashline_reset();
	for (new = root; new; new = new->l_chain)
	{
		if (str_equal(filename, new->l_name))
		{
			lex_error("%S: recursive include", filename);
			return;
		}
	}
	new = mem_alloc_clear(sizeof(lex_ty));
	new->l_name = str_copy(filename);
	new->l_file = fopen(filename->str_text, "r");
	if (!new->l_file)
		nfatal("cookbook \"%S\"", filename);
	new->l_line = 1;
	new->l_bol = 1;
	new->l_chain = root;
	wl_zero(&new->pending_includes);
	root = new;
	trace((/*{*/"}\n"));
}


void
lex_open_include(s)
	string_ty	*s;
{
	assert(root);
	wl_append(&root->pending_includes, s);
}


/*
 * NAME
 *	lex_close - closes a file which was open for lexical analysis
 *
 * SYNOPSIS
 *	void lex_close(void);
 *
 * DESCRIPTION
 *	Lex_close closes a file previously opened for lexical analysis.
 *
 * CAVEAT
 *	Lex_open must have been previously used to open the file.
 */

void
lex_close()
{
	lex_ty		*old;

	trace(("lex_close()\n{\n"/*}*/));
	assert(root);
	old = root;
	fclose(root->l_file);
	root = old->l_chain;
	if (!root && errcnt)
		exit(1);
	str_free(old->l_name);
	wl_free(&old->pending_includes);
	free(old);
	trace((/*{*/"}\n"));
}


/*
 * NAME
 *	illeof - conplain about an illegal end-of-file
 *
 * SYNOPSIS
 *	void illeof(void);
 *
 * DESCRIPTION
 *	Illeof is used to complain about illegal end-of-file.
 *	This may occu in several places.
 *
 * CAVEAT
 *	It does not return.
 */

static void illeof _((void));

static void
illeof()
{
	lex_error("unexpected end of file");
	exit(1);
}


/*
 * NAME
 *	byte - get a byte from the input stream
 *
 * SYNOPSIS
 *	int byte(void);
 *
 * DESCRIPTION
 *	Byte return the next byte from the input stream, or EOF
 *	If the end of the outermost file is reached.
 *	It detechs the end of include files and closes them transparently.
 *
 * CAVEAT
 *	Lex_open must have been called previously.
 */

static int byte _((void));

static int
byte()
{
	int		c;
		
	trace(("byte()\n{\n"/*}*/));
	for (;;)
	{
		/*
		 * leap into include file if possible
		 */
		assert(root);
		if (root->pending_includes.wl_nwords)
		{
			string_ty	*fn;

			fn = str_copy(root->pending_includes.wl_word[0]);
			wl_delete(&root->pending_includes, fn);
			lex_open(fn);
			str_free(fn);
			star();
		}

		/*
		 * fetch the next input character
		 */
		c = fgetc(root->l_file);
		if (c != EOF)
			break;
		if (ferror(root->l_file))
			nfatal("read %S", root->l_name);
		if (!root->l_chain)
			break;
		lex_close();
	}
	if (!isprint(c) && c != EOF && !strchr("\t\n\f", c))
	{
		lex_error("illegal '\\%o' character", (unsigned char)c);
		c = ' ';
	}
	if (mode != LM_DATA && root->l_bol && c == '#')
		c = HASHLINE_ESCAPE;
	if (c == '\n')
	{
		root->l_bol = 1;
		root->l_line++;
	}
	else
	{
		if (!strchr(" \t\f", c))
			root->l_bol = 0;
	}
	trace_short(root->l_bol);
	trace(("return '%c';\n", c));
	trace((/*{*/"}\n"));
	return c;
}


/*
 * NAME
 *	byte_undo - push back a character
 *
 * SYNOPSIS
 *	void byte_undo(int c);
 *
 * Description
 *	Byte_undo pushes a byte back onto the input stream.
 *
 * CAVEAT
 *	Only one byte may be pushed back at any one time.
 */

static void byte_undo _((int));

static void
byte_undo(c)
	int		c;
{
	trace(("byte_undo(c = '%c')\n{\n"/*}*/, c));
	assert(root);
	switch (c)
	{
	case EOF:
		break;

	case '\n':
		root->l_line--;
	default:
		ungetc(c, root->l_file);
		break;
	}
	trace((/*{*/"}\n"));
}


/*
 * NAME
 *	sa_open - start accumulating a string.
 *
 * SYNOPSIS
 *	void sa_open(void);
 *
 * DESCRIPTION
 *	Sa_open begins accumulating a string for the lexical analyser.
 *	This allows virtually infinite length constructs within the
 *	lexical analyser.
 */

static void sa_open _((void));

static void
sa_open()
{
	trace(("sa_open()\n{\n"/*}*/));
	assert(!saroot.sa_inuse);
	saroot.sa_inuse = 1;
	saroot.sa_len = 0;
	trace((/*{*/"}\n"));
}


/*
 * NAME
 *	sa_char - add a character to the accumulating string
 *
 * SYNOPSIS
 *	void sa_char(int c);
 *
 * DESCRIPTION
 *	Sa_char adds a character to the accumulating string.
 *
 * CAVEAT
 *	Sa_open must have been called previously.
 */

static void sa_char _((int));

static void
sa_char(c)
	int		c;
{
	trace(("sa_char(c = '%c')\n{\n"/*}*/, c));
	assert(saroot.sa_inuse);
	if (saroot.sa_len >= saroot.sa_max)
	{
		saroot.sa_max += 128;
		saroot.sa_buf = mem_change_size(saroot.sa_buf, saroot.sa_max);
	}
	saroot.sa_buf[saroot.sa_len++] = c;
	trace((/*{*/"}\n"));
}


/*
 * NAME
 *	sa_close - finish accumulating a string
 *
 * SYNOPSIS
 *	string_ty *sa_close(void);
 *
 * DESCRIPTION
 *	Sa_close finished accumulating a string and
 *	returns a pointer to the string in dynamic memory.
 *
 * CAVEAT
 *	Sa_open must have been called previously.
 *
 *	The value returned by this function is allocated in dynamic memory.
 *	It is the responsibility of the caller to ensure that it is freed when
 *	finished with, by a call to str_free().
 */

static string_ty *sa_close _((void));

static string_ty *
sa_close()
{
	string_ty	*val;

	trace(("sa_close()\n{\n"/*}*/));
	assert(saroot.sa_inuse);
	val = str_n_from_c(saroot.sa_buf, saroot.sa_len);
	saroot.sa_inuse = 0;
	trace(("return %08lX;\n", val));
	trace((/*{*/"}\n"));
	return val;
}


/*
 * NAME
 *	sa_mark - mark place when accumulating a string
 *
 * SYNOPSIS
 *	size_t sa_mark(void);
 *
 * DESACRIPTION
 *	Sa_mark returns an indicator as to where the string being accumulated
 *	is up to.
 *
 * CAVEAT
 *	Sa_open must have been called previously.
 *	Do not use the returned value for anything other than passing to
 *	sa_goto as an argument.
 */

static size_t sa_mark _((void));

static size_t
sa_mark()
{
	assert(saroot.sa_inuse);
	return (saroot.sa_len);
}


/*
 * NAME
 *	sa_goto - shorten the accumulated string
 *
 * SYNOPSIS
 *	void sa_goto(size_t mark);
 *
 * DESACRIPTION
 *	Sa_goto shotens the accumulating string to a point
 *	previously determined by sa_mark.
 *
 * CAVEAT
 *	Sa_open must have been called previously.
 *	The mark argument must have previously been returned by sa_mark,
 *	and sa_close not called in the interim.
 */

static void sa_goto _((size_t));

static void
sa_goto(n)
	size_t		n;
{
	assert(saroot.sa_inuse);
	assert(n <= saroot.sa_len);
	saroot.sa_len = n;
}


/*
 * NAME
 *	meta_repn - mata_ty representation
 *
 * SYNOPSIS
 *	char *meta_repn(meta_ty*);
 *
 * DESCRIPTION
 *	The meta_repn function is used to produce a readable representation of
 *	a mata_ty value.
 *
 * RETURNS
 *	char* - a pointer to ta C string.
 *
 * CAVEAT
 *	This function is only available when the DEBUG symbol is #define'd.
 */

#ifdef DEBUG

static char *meta_repn _((meta_ty *));

static char *
meta_repn(val)
	meta_ty		*val;
{
	static char *type[] =
	{
		"0",
		"NORMAL",
		"SLOSHED",
		"WHITE",
	};
	static char *flag[] =
	{
		"0",
		"BEFORE",
		"AFTER",
		"AFTER | BEFORE",
		"SINGLE",
		"SINGLE | BEFORE",
		"SINGLE | AFTER",
		"SINGLE | AFTER | BEFORE",
	};
	static char	buffer[100];
	char		buf2[20];
	char		buf3[10];

	if (val->m_type >= 0 && val->m_type < SIZEOF(type))
		strcpy(buf2, type[val->m_type]);
	else
		sprintf(buf2, "%d", val->m_type);
	switch (val->m_char)
	{
	default:
		if (isprint(val->m_char))
		{
			buf3[0] = val->m_char;
			buf3[1] = 0;
		}
		else
			sprintf(buf3, "\\%03o", (unsigned char)val->m_char);
		break;

	case '\'':
	case '\\':
		buf3[0] = '\\';
		buf3[1] = val->m_char;
		buf3[2] = 0;
		break;
	}
	sprintf
	(
		buffer,
		"{m_char = '%s', m_type = %s, m_flag = %s}",
		buf3,
		buf2,
		flag[val->m_flag]
	);
	return buffer;
}

#endif


/*
 * NAME
 *	meta - fetch and classify a character
 *
 * SYNOPSIS
 *	void meta(meta_ty*);
 *
 * DESCRIPTION
 *	Meta fetches and classifies a character from the input stream.
 *	The classification may be 'mode' dependent.
 *
 *	This is the point at which escaped newlines and comments are replaced
 *	by a single space.
 *
 * CAVEAT
 *	Lex_open must have been called previously.
 */

static void meta _((meta_ty *));

static void
meta(val)
	meta_ty		*val;
{
	trace(("meta(val = %08lX)\n{\n"/*}*/, val));
	assert(root);
	if (root->l_mback.m_type)
	{
		*val = root->l_mback;
		root->l_mback.m_type = 0;
	}
	else
	{
		switch (val->m_char = byte())
		{
		default:
			val->m_type = NORMAL;
			val->m_flag = BEFORE | AFTER;
			break;
	
		case EOF:
			val->m_type = EOF;
			val->m_flag = SINGLE;
			break;
	
		case '=':
		case ';':
		case ':':
		case '{'/*}*/:
		case /*{*/'}':
			switch (mode)
			{
			default:
				val->m_type = NORMAL;
				val->m_flag = SINGLE;
				break;
	
			case LM_DATA:
			case LM_SQUOTE:
			case LM_DQUOTE:
				val->m_type = NORMAL;
				val->m_flag = BEFORE | AFTER;
				break;
			}
			break;
	
		case '['/*]*/:
			switch (mode)
			{
			default:
				val->m_flag = BEFORE | SINGLE;
				break;
	
			case LM_SQUOTE:
			case LM_DQUOTE:
				val->m_flag = BEFORE | AFTER;
			}
			val->m_type = NORMAL;
				break;
	
		case /*[*/']':
			val->m_type = NORMAL;
			switch (mode)
			{
			default:
				val->m_flag = AFTER | SINGLE;
				break;
	
			case LM_DATA:
			case LM_SQUOTE:
			case LM_DQUOTE:
				val->m_flag = BEFORE | AFTER;
				break;
			}
			break;
	
		case ' ':
		case '\t':
		case '\n':
		case '\f':
			switch (mode)
			{
			default:
				val->m_type = WHITE;
				val->m_flag = SINGLE;
				break;
	
			case LM_SQUOTE:
			case LM_DQUOTE:
			case LM_DATA:
				val->m_type = NORMAL;
				val->m_flag = BEFORE | AFTER;
				break;
			}
			break;
	
		case '\\':
			switch (val->m_char = byte())
			{
			case EOF:
				illeof();
	
			case '\n':
				val->m_char = ' ';
				val->m_type = WHITE;
				val->m_flag = SINGLE;
				break;
	
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
				{
					int	 count;
					int	 c;
	
					c = val->m_char;
					val->m_char = 0;
					for (count = 0; count < 3; count++)
					{
						val->m_char = val->m_char * 8 + c - '0';
						c = byte();
						if (c < '0' || c > '7')
						{
							byte_undo(c);
							break;
						}
					}
					val->m_type = SLOSHED;
					val->m_flag = BEFORE | AFTER;
				}
				break;
	
			default:
				{
					char	*cp;
	
					cp = strchr("b\bf\fn\nr\rt\t", val->m_char);
					if (cp)
						val->m_char = cp[1];
					val->m_type = SLOSHED;
					val->m_flag = BEFORE | AFTER;
				}
				break;
			}
			break;
	
		case '/':
			switch (mode)
			{
			default:
				{
					int	 c;
					int	 gate;
					int	 count;
	
					if ((c = byte()) != '*')
					{
						byte_undo(c);
						val->m_type = NORMAL;
						val->m_flag = BEFORE | AFTER;
						break;
					}
					gate = 0;
					count = 1;
					while (count)
					{
						switch (c = byte())
						{
						default:
						case '\n':
						case '\t':
						case '\f':
							gate = 0;
							break;
	
						case EOF:
							illeof();
	
						case '*':
							if (gate == 1)
							{
								/*
								 * nested comment start
								 *
								 * should a warning be issued?
								 */
								count++;
								gate = 0;
							}
							else
								gate = 2;
							break;
	
						case '/':
							if (gate == 2)
							{
								count--;
								gate = 0;
							}
							else
								gate = 1;
							break;
						}
					}
					val->m_char = ' ';
					val->m_type = WHITE;
					val->m_flag = SINGLE;
				}
				break;
	
			case LM_DATA:
			case LM_SQUOTE:
			case LM_DQUOTE:
				val->m_type = NORMAL;
				val->m_flag = BEFORE | AFTER;
				break;
			}
			break;
		}
	}
	trace(("*val = %s;\n", meta_repn(val)));
	trace((/*{*/"}\n"));
}


/*
 * NAME
 *	meta_undo - push back a character
 *
 * SYNOPSIS
 *	void meta_undo(meta_ty*);
 *
 * DESCRIPTION
 *	Meta_undo is used to temporarily push back a character
 *	from the input stream.
 *
 * CAVEAT
 *	Only one character may be pushed back at any one time.
 */

static void meta_undo _((meta_ty *));

static void
meta_undo(val)
	meta_ty		*val;
{
	assert(root);
	assert(!root->l_mback.m_type);
	root->l_mback = *val;
}


/*
 * NAME
 *	lex_error - print a file location specific error message
 *
 * SYNOPSIS
 *	void lex_error(char *s, ...);
 *
 * DESCRIPTION
 *	Lex_error prints a file location specific error message.
 *	If too many errors pass through here, cook will give up.
 */

void
lex_error(s sva_last)
	char		*s;
	sva_last_decl
{
	string_ty	*buffer;
	va_list		ap;

	sva_init(ap, s);
	assert(root);
	buffer = str_vformat(s, ap);
	va_end(ap);
	error("%S: %d: %S", root->l_name, root->l_line, buffer);
	str_free(buffer);
	if (++errcnt > 20)
		fatal("%S: too many errors", root->l_name);
	option_set_errors();
}


/*
 * NAME
 *	lex_warning - print a file location specific warning message
 *
 * SYNOPSIS
 *	void lex_warning(char *s, ...);
 *
 * DESCRIPTION
 *	Lex_warning prints a file location specific warning message.
 */

void
lex_warning(s sva_last)
	char		*s;
	sva_last_decl
{
	string_ty	*buffer;
	va_list		ap;

	assert(root);
	sva_init(ap, s);
	buffer = str_vformat(s, ap);
	va_end(ap);
	error("%S: %d: warning: %S", root->l_name, root->l_line, buffer);
	str_free(buffer);
}


/*
 * NAME
 *	lex_mode - change the mode of lexical analysis
 *
 * SYNOPSIS
 *	int lex_mode(int n);
 *
 * DESCRIPTION
 *	Lex_mode changes the mode of lexical analysis to the one given,
 *	and returns the previous mode.
 *
 * CAVEAT
 *	Use the #define's in lex.h.
 */

int
lex_mode(n)
	int		n;
{
	int		old;

	old = mode;
	mode = n;
	return (old);
}


/*
 * NAME
 *	lex_trace - debug output
 *
 * SYNOPSIS
 *	void lex_trace(char *s, ...);
 *
 * DESCRIPTION
 *	The lex_trace function is used to format and output yyparse's trace
 *	output.  The printf's in yyparse are #define'd into lex_trace calls.
 *	Unfortunately yacc's designer did not take trace output redirection
 *	into account, to this function is a little tedious.
 *
 * RETURNS
 *	void
 *
 * CAVEAT
 *	This function is only available when the DEBUG symbol is #define'd.
 */

#ifdef DEBUG

/*VARARGS1*/
void
lex_trace(s sva_last)
	char		*s;
	sva_last_decl
{
	va_list		ap;
	string_ty	*buffer;
	char		*cp;
	static char	line[1024];

	sva_init(ap, s);
	buffer = str_vformat(s, ap);
	va_end(ap);
	strcat(line, buffer->str_text);
	str_free(buffer);
	cp = line + strlen(line) - 1;
	if (cp > line && *cp == '\n')
	{
		*cp = 0;
		trace_printf("%S: %d: %s\n", root->l_name, root->l_line, line);
		line[0] = 0;
	}
}


static char *mode_repn _((int));

static char *
mode_repn(n)
	int	n;
{
	static char *name[] =
	{
		"NORMAL",
		"DATA",
		"SQUOTE",
		"DQUOTE",
	};
	static char buffer[12];

	if (n >= 0 && n < SIZEOF(name))
		return name[n];
	sprintf(buffer, "%d", n);
	return buffer;
}

#endif


/*
 * NAME
 *	tokenize - lexical analyser
 *
 * SYNOPSIS
 *	int tokenize(void);
 *
 * DESCRIPTION
 *	The tokenize functionm is used to partition the input stream into
 *	discreet tokens. 
 *
 * CAVEAT
 *	Lex_open must have been called previously.
 */

static int tokenize _((void));

static int
tokenize()
{
	meta_ty		c;
	int		tok;
	static meta_ty	last;
	static int	dataended;

	trace(("tokenize()\n{\n"/*}*/));
	if (dataended)
	{
		dataended = 0;
		tok = DATAEND;
		goto ret;
	}

	if (!catted && (last.m_flag & AFTER))
	{
		/*
		 * For parts of words in expressions which are abutted
		 * together, the user is implying catenation of the strings.
		 * This implied zero length token happens here.
		 */
		meta(&c);
		meta_undo(&c);
		if (c.m_flag & BEFORE)
		{
			catted = 1;
			tok = CATENATE;
			goto ret;
		}
	}
	else
		catted = 0;

	for (;;)
	{
		int	 mark;

		mark = 0;
		meta(&c);
		if (c.m_char == HASHLINE_ESCAPE)
		{
			state = 0;
			last.m_flag = 0;
			hashline();
			state = 0;
			last.m_flag = 0;
			continue;
		}
		switch (c.m_type)
		{
		default:
			fatal("illegal m_type %d (bug %d)", c.m_type, __LINE__);

		case EOF:
			tok = 0; /* by definition of yacc */
			goto ret;

		case NORMAL:
		case SLOSHED:
			{
				int	 sloshed;

				sloshed = 0;
				sa_open();
				if (state == 0)
					mark = sa_mark();
				for (;;)
				{
					trace
					((
							"%s: %d: state %d, mode %s;\n",
						root->l_name->str_text,
						root->l_line,
						state,
						mode_repn(mode)
					));
					sa_char(c.m_char);
					if (c.m_type == SLOSHED)
					{
						sloshed = 1;
						switch (state)
						{
						default:
							state = 1;
							break;

						case 11:
						case 12:
							/* don't change state */
							break;
						}
					}
					else
					{
						switch (c.m_char)
						{
						default:
							switch (state)
							{
							default:
								state = 1;

							case 11:
							case 12:
								break;
							}
							break;

						case '\n':
							switch (state)
							{
							case 9:
							case 10:
								sa_goto(mark);
								parse_lval.lv_word = sa_close();
								state = 0;
								last.m_flag = 0; /* don't CAT after this */
								dataended = DATAEND;
								tok = WORD;
								goto ret;

							case 11:
							case 12:
								/* don't change state */
								lex_error
								(
							      "newline in quote"
								);
								break;

							default:
								state = 0;
								mark = sa_mark();
								break;
							}
							break;

						case '\'':
							if (state != 12)
							{
								static		int		oldmode = LM_SQUOTE;

								oldmode = lex_mode(oldmode);
								sa_goto(sa_mark()-1);
								state = ((state == 11) ? 1 : 11);
								sloshed = 1;
							}
							break;

						case '"':
							if (state != 11)
							{
								static		int		oldmode = LM_DQUOTE;

								oldmode = lex_mode(oldmode);
								sa_goto(sa_mark()-1);
								state = ((state == 12) ? 1 : 12);
								sloshed = 1;
							}
							break;

						case ' ':
						case '\t':
							switch (state)
							{
							default:
								state = 1;
								break;

							case 0:
							case 2:
								state = 2;
								break;

							case 9:
							case 10:
								state = 10;
								break;

							case 11:
							case 12:
								break;
							}
							break;

						case 'a':
							switch (state)
							{
							default:
								state = 1;
								break;

							case 3:
							case 5:
								state++;
								break;

							case 11:
							case 12:
								break;
							}
							break;

						case 'd':
							switch (state)
							{
							default:
								state = 1;
								break;

							case 0:
							case 2:
								state = 3;
								break;

							case 8:
								state = 9;
								break;

							case 11:
							case 12:
								break;
							}
							break;

						case 'e':
							switch (state)
							{
							default:
								state = 1;
								break;

							case 6:
								state = 7;
								break;

							case 11:
							case 12:
								break;
							}
							break;

						case 'n':
							switch (state)
							{
							default:
								state = 1;
								break;

							case 7:
								state = 8;
								break;

							case 11:
							case 12:
								break;
							}
							break;

						case 't':
							switch (state)
							{
							default:
								state = 1;
								break;

							case 4:
								state = 5;
								break;

							case 11:
							case 12:
								break;
							}
							break;
						}
					}
					last = c;
					if (last.m_flag & SINGLE)
						break;
					meta(&c);
					if (c.m_flag & SINGLE)
					{
						meta_undo(&c);
						break;
					}
				}
				parse_lval.lv_word = sa_close();
				if (sloshed)
					tok = WORD;
				else
					tok = UNQUOTED_WORD;
			}
			goto ret;

		case WHITE:
			if (c.m_char == '\n')
			{
				last = c;
				state = 0;
				tok = EOLN;
				goto ret;
			}
			state = 1;
			break;
		}
	}

	ret:
	trace
	((
		"%S: %d: state %d, mode %s;\n",
		root->l_name,
		root->l_line,
		state,
		mode_repn(mode)
	));
	trace(("return %d;\n", tok));
	trace((/*{*/"}\n"));
	return tok;
}


/*
 * NAME
 *	lex_cur_file - name current file
 *
 * SYNOPSIS
 *	string_ty *lex_cur_file(void);
 *
 * DESCRIPTION
 *	The lex_cur_file function is used to get the name of the current
 *	file being analyzed by lex.
 *
 * RETURNS
 *	string_ty* - pointer to string with name in it
 *
 * CAVEAT
 *	Does not take a copy, so don't use str_free()
 */

string_ty *
lex_cur_file()
{
	assert(root);
	return root->l_name;
}


int
lex_cur_line()
{
	assert(root);
	return root->l_line;
}


/*
 * NAME
 *	parse_lex - lexer for parse.y
 *
 * SYNOPSIS
 *	int parse_lex(void);
 *
 * DESCRIPTION
 *	The parse_lex function is used to turn the tokens generated
 *	by tokenize into parse.y tokens.
 *
 * RETURNS
 *	int - the token number
 *
 * CAVEAT
 *	Intended solely for use by the yacc-generated compiler
 */

int
parse_lex()
{
	int		tok;
	int		*data;

	trace(("parse_lex()\n{\n"/*}*/));
	for (;;)
	{
		tok = tokenize();
		if (!passing)
		{
			if (!tok)
			{
				lex_error("unterminated conditional");
				break;
			}
			if (tok == WORD || tok == UNQUOTED_WORD)
				str_free(parse_lval.lv_word);
			continue;
		}
		switch (tok)
		{
		case EOLN:
			continue;

		case UNQUOTED_WORD:
			assert(parse_symtab);
			data = symtab_query(parse_symtab, parse_lval.lv_word);
			if (data)
			{
				tok = *data;
				str_free(parse_lval.lv_word);
				parse_lval.lv_word = 0;
				if (tok == COLON)
				{
					parse_lval.lv_position.pos_name = str_copy(root->l_name);
					parse_lval.lv_position.pos_line = root->l_line;
				}
				if (tok == DATA)
				{
					int	winged;
					meta_ty	last;
					meta_ty	c;
	
					/*
					 * This is a special.  Gobble the rest
					 * of the line before he starts sucking
					 * in the program data.
					 */
					winged = 0;
					for (;;)
					{
						meta(&c);
						last = c;
						switch (c.m_type)
						{
						default:
							fatal
							(
						   "illegal m_type %d (bug %d)",
								c.m_type,
								__LINE__
							);
	
						case EOF:
							illeof();
	
						case NORMAL:
						case SLOSHED:
							if (!winged)
							{
								lex_error
								(
			       "the 'data' keyword must be the last on the line"
								);
								winged = 1;
							}
							continue;
	
						case WHITE:
							if (c.m_char != '\n')
								continue;
							state = 0;
							break;
						}
						break;
					}
					catted = 1;
				}
			}
			else
				tok = WORD;
			break;
		}
		break;
	}
	trace(("return %d;\n", tok));
	trace((/*{*/"}\n"));
	return tok;
}


/*
 * NAME
 *	hashline_lex - lexer for hashline.y
 *
 * SYNOPSIS
 *	int hashline_lex(void);
 *
 * DESCRIPTION
 *	The hashline_lex function is used to turn the tokens generated
 *	by tokenize into hashline.y tokens.
 *
 * RETURNS
 *	int - the token number
 *
 * CAVEAT
 *	Intended solely for use by the yacc-generated compiler
 */

int
hashline_lex()
{
	int		tok;
	int		*data;

	trace(("hashline_lex()\n{\n"/*}*/));
	tok = tokenize();
	switch (tok)
	{
	case EOLN:
		tok = 0;
		break;

	case UNQUOTED_WORD:
		assert(hash_symtab);
		data = symtab_query(hash_symtab, parse_lval.lv_word);
		if (data)
		{
			tok = *data;
			str_free(parse_lval.lv_word);
			parse_lval.lv_word = 0;
		}
		else
			tok = WORD;
		break;
	}
	memcpy(&hashline_lval, &parse_lval, sizeof(hashline_lval));
	trace(("return %d;\n", tok));
	trace((/*{*/"}\n"));
	return tok;
}


void
lex_passing(n)
	int		n;
{
	passing = n;
}
